///--------------------------------------------------------------------------------------
//	SpriteLayer.c
//
//	Portions are copyright:  1991-94 Tony Myles, All rights reserved worldwide.
//
//	Description:	implementation of the sprite layers
///--------------------------------------------------------------------------------------


#ifndef __SWCOMMON__
#include "SWCommonHeaders.h"
#endif

#ifndef __MEMORY__
#include <Memory.h>
#endif

#ifndef __SPRITEWORLDUTILS__
#include "SpriteWorldUtils.h"
#endif

#ifndef __SPRITEWORLD__
#include "SpriteWorld.h"
#endif

#ifndef __SPRITELAYER__
#include "SpriteLayer.h"
#endif

#ifndef __SPRITE__
#include "Sprite.h"
#endif



///--------------------------------------------------------------------------------------
//	SWCreateSpriteLayer
///--------------------------------------------------------------------------------------

SW_FUNC OSErr SWCreateSpriteLayer(
	SpriteLayerPtr *spriteLayerP)
{
	OSErr err;
	SpriteLayerPtr tempSpriteLayerP;

	err = noErr;
	*spriteLayerP = NULL;

	tempSpriteLayerP = (SpriteLayerPtr)NewPtrClear((Size)sizeof(SpriteLayerRec));

	if (tempSpriteLayerP != NULL)
	{
		*spriteLayerP = tempSpriteLayerP;
	}
	else
	{
		err = MemError();
	}

	SWSetStickyIfError( err );
	return err;
}


///--------------------------------------------------------------------------------------
//	SWDisposeSpriteLayer
///--------------------------------------------------------------------------------------

SW_FUNC void SWDisposeSpriteLayer(
	SpriteLayerPtr spriteLayerP)
{
	if (spriteLayerP != NULL)
	{
		DisposePtr((Ptr)spriteLayerP);
	}
}


///--------------------------------------------------------------------------------------
//	SWAddSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWAddSprite(
	SpriteLayerPtr spriteLayerP,
	SpritePtr newSpriteP)
{
	SpritePtr tailSpriteP = spriteLayerP->tailSpriteP;

	if (tailSpriteP != NULL)
	{
			// doubly link the new sprite
		tailSpriteP->nextSpriteP = newSpriteP;
		newSpriteP->prevSpriteP = tailSpriteP;
		newSpriteP->nextSpriteP = NULL;
	}	
	else
	{
		newSpriteP->prevSpriteP = NULL;
		newSpriteP->nextSpriteP = NULL;

			// make the new sprite the head
		spriteLayerP->headSpriteP = newSpriteP;
	}

		// make the new sprite the tail
	spriteLayerP->tailSpriteP = newSpriteP;
	
	newSpriteP->spriteRemoval = kSWDontRemoveSprite;
}


///--------------------------------------------------------------------------------------
//	SWRemoveSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWRemoveSprite(
	SpriteLayerPtr spriteLayerP,
	SpritePtr oldSpriteP)
{
		// is this not the tail sprite?
	if (oldSpriteP->nextSpriteP != NULL)
	{
			// link the next sprite to the prev sprite
		oldSpriteP->nextSpriteP->prevSpriteP = oldSpriteP->prevSpriteP;
	}
	else
	{
			// make the prev sprite the tail
		spriteLayerP->tailSpriteP = oldSpriteP->prevSpriteP;
	}

		// is this not the head sprite?
	if (oldSpriteP->prevSpriteP != NULL)
	{
			// link the prev sprite to the next sprite
		oldSpriteP->prevSpriteP->nextSpriteP = oldSpriteP->nextSpriteP;
	}
	else
	{
			// make the next sprite the first sprite
		spriteLayerP->headSpriteP = oldSpriteP->nextSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWRemoveAllSpritesFromLayer
///--------------------------------------------------------------------------------------

SW_FUNC void SWRemoveAllSpritesFromLayer(
	SpriteLayerPtr srcSpriteLayerP)
{
	SpritePtr curSpriteP;

	while ((curSpriteP = SWGetNextSprite(srcSpriteLayerP, NULL)) != NULL)
	{
		SWRemoveSprite(srcSpriteLayerP, curSpriteP);
	}
}


///--------------------------------------------------------------------------------------
//	SWSwapSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWSwapSprite(
	SpriteLayerPtr spriteLayerP,
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP)
{
	register SpritePtr swapSpriteP;
	
	
		// adjacent Sprites are a special case
		
	if ( srcSpriteP->nextSpriteP == dstSpriteP ||
		dstSpriteP->nextSpriteP == srcSpriteP )
	{
		if ( srcSpriteP->nextSpriteP == dstSpriteP )
		{
			if ( srcSpriteP->prevSpriteP != NULL )
				(srcSpriteP->prevSpriteP)->nextSpriteP = dstSpriteP;
			if ( dstSpriteP->nextSpriteP != NULL )
				(dstSpriteP->nextSpriteP)->prevSpriteP = srcSpriteP;
			
			dstSpriteP->prevSpriteP = srcSpriteP->prevSpriteP;
			srcSpriteP->nextSpriteP = dstSpriteP->nextSpriteP;
			
			dstSpriteP->nextSpriteP = srcSpriteP;
			srcSpriteP->prevSpriteP = dstSpriteP;
		}
		else
		{
			if ( dstSpriteP->prevSpriteP != NULL )
				(dstSpriteP->prevSpriteP)->nextSpriteP = srcSpriteP;
			if ( srcSpriteP->nextSpriteP != NULL )
				(srcSpriteP->nextSpriteP)->prevSpriteP = dstSpriteP;
			
			srcSpriteP->prevSpriteP = dstSpriteP->prevSpriteP;
			dstSpriteP->nextSpriteP = srcSpriteP->nextSpriteP;
			
			srcSpriteP->nextSpriteP = dstSpriteP;
			dstSpriteP->prevSpriteP = srcSpriteP;
		}
	}
	else
	{
		if ( srcSpriteP->prevSpriteP != NULL && dstSpriteP->prevSpriteP != NULL )
		{
			swapSpriteP = (srcSpriteP->prevSpriteP)->nextSpriteP;
			(srcSpriteP->prevSpriteP)->nextSpriteP = 
					(dstSpriteP->prevSpriteP)->nextSpriteP;
			(dstSpriteP->prevSpriteP)->nextSpriteP = swapSpriteP;
		}
		if ( srcSpriteP->nextSpriteP != NULL && dstSpriteP->nextSpriteP != NULL )
		{
			swapSpriteP = (srcSpriteP->nextSpriteP)->prevSpriteP;
			(srcSpriteP->nextSpriteP)->prevSpriteP = 
					(dstSpriteP->nextSpriteP)->prevSpriteP;
			(dstSpriteP->nextSpriteP)->prevSpriteP = swapSpriteP;
		}
		
		swapSpriteP = srcSpriteP->nextSpriteP;
		srcSpriteP->nextSpriteP = dstSpriteP->nextSpriteP;
		dstSpriteP->nextSpriteP = swapSpriteP;

		swapSpriteP = srcSpriteP->prevSpriteP;
		srcSpriteP->prevSpriteP = dstSpriteP->prevSpriteP;
		dstSpriteP->prevSpriteP = swapSpriteP;
	}


	if (srcSpriteP->nextSpriteP == NULL)
	{
		spriteLayerP->tailSpriteP = srcSpriteP;
	}
	else if (srcSpriteP->prevSpriteP == NULL)
	{
		spriteLayerP->headSpriteP = srcSpriteP;
	}

	if (dstSpriteP->nextSpriteP == NULL)
	{
		spriteLayerP->tailSpriteP = dstSpriteP;
	}
	else if (dstSpriteP->prevSpriteP == NULL)
	{
		spriteLayerP->headSpriteP = dstSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWInsertSpriteAfterSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWInsertSpriteAfterSprite(
	SpriteLayerPtr spriteLayerP,
	SpritePtr newSpriteP,
	SpritePtr dstSpriteP)
{

	if ( dstSpriteP->nextSpriteP != NULL )
		(dstSpriteP->nextSpriteP)->prevSpriteP = newSpriteP;
	
	newSpriteP->nextSpriteP = dstSpriteP->nextSpriteP;
	dstSpriteP->nextSpriteP = newSpriteP;
	newSpriteP->prevSpriteP = dstSpriteP;
	
	if (newSpriteP->nextSpriteP == NULL)
	{
		spriteLayerP->tailSpriteP = newSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWInsertSpriteBeforeSprite
///--------------------------------------------------------------------------------------

SW_FUNC void SWInsertSpriteBeforeSprite(
	SpriteLayerPtr spriteLayerP,
	SpritePtr newSpriteP,
	SpritePtr dstSpriteP)
{

	if ( dstSpriteP->prevSpriteP != NULL )
		(dstSpriteP->prevSpriteP)->nextSpriteP = newSpriteP;
	
	newSpriteP->prevSpriteP = dstSpriteP->prevSpriteP;
	dstSpriteP->prevSpriteP = newSpriteP;
	newSpriteP->nextSpriteP = dstSpriteP;
	
	if (newSpriteP->prevSpriteP == NULL)
	{
		spriteLayerP->headSpriteP = newSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWGetNextSprite
///--------------------------------------------------------------------------------------

SW_FUNC SpritePtr SWGetNextSprite(
	SpriteLayerPtr spriteLayerP,
	SpritePtr curSpriteP)
{
	return (curSpriteP == NULL) ? spriteLayerP->headSpriteP : curSpriteP->nextSpriteP;
}


///--------------------------------------------------------------------------------------
//	SWLockSpriteLayer
///--------------------------------------------------------------------------------------

SW_FUNC void SWLockSpriteLayer(
	SpriteLayerPtr spriteLayerP)
{
	SpritePtr curSpriteP;

	curSpriteP = spriteLayerP->headSpriteP;

	while (curSpriteP != NULL)
	{
		SWLockSprite(curSpriteP);

		curSpriteP = curSpriteP->nextSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWUnlockSpriteLayer
///--------------------------------------------------------------------------------------

SW_FUNC void SWUnlockSpriteLayer(
	SpriteLayerPtr spriteLayerP)
{
	SpritePtr curSpriteP;

	curSpriteP = spriteLayerP->headSpriteP;

	while (curSpriteP != NULL)
	{
		SWUnlockSprite(curSpriteP);

		curSpriteP = curSpriteP->nextSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWCollideSpriteLayer
///--------------------------------------------------------------------------------------

SW_FUNC void SWCollideSpriteLayer(
	SpriteWorldPtr	spriteWorldP,
	SpriteLayerPtr srcSpriteLayerP,
	SpriteLayerPtr dstSpriteLayerP)
{
	SpritePtr srcSpriteP, nextSrcSpriteP;
	SpritePtr dstSpriteP, nextDstSpriteP;
	Rect sectRect;
	
		// Don't check for collisions unless the frame has been processed!
	if (!spriteWorldP->frameHasOccured)
		return;

	srcSpriteP = srcSpriteLayerP->headSpriteP;

	while (srcSpriteP != NULL)
	{
		dstSpriteP = dstSpriteLayerP->headSpriteP;
		nextSrcSpriteP = srcSpriteP->nextSpriteP;

		while (dstSpriteP != NULL)
		{
			nextDstSpriteP = dstSpriteP->nextSpriteP;

			if (srcSpriteP != dstSpriteP)
			{
					// are the sprites rectangles overlapping?
				if ((srcSpriteP->destFrameRect.top < dstSpriteP->destFrameRect.bottom) &&
					(srcSpriteP->destFrameRect.bottom > dstSpriteP->destFrameRect.top) &&
					(srcSpriteP->destFrameRect.left < dstSpriteP->destFrameRect.right) &&
					(srcSpriteP->destFrameRect.right > dstSpriteP->destFrameRect.left))
				{
						// call the source sprites collision routine
					if (srcSpriteP->spriteCollideProc != NULL)
					{
						sectRect.left =
							SW_MAX(srcSpriteP->destFrameRect.left, dstSpriteP->destFrameRect.left);
						sectRect.top =
							SW_MAX(srcSpriteP->destFrameRect.top, dstSpriteP->destFrameRect.top);
						sectRect.right =
							SW_MIN(srcSpriteP->destFrameRect.right, dstSpriteP->destFrameRect.right);
						sectRect.bottom =
							SW_MIN(srcSpriteP->destFrameRect.bottom, dstSpriteP->destFrameRect.bottom);

						(*srcSpriteP->spriteCollideProc)(srcSpriteP, dstSpriteP, &sectRect);
					}
				}
			}
			
			dstSpriteP = nextDstSpriteP;
		}
		
		srcSpriteP = nextSrcSpriteP;
	}
}


///--------------------------------------------------------------------------------------
//	SWFindSpriteByPoint
///--------------------------------------------------------------------------------------

SpritePtr SWFindSpriteByPoint(
	SpriteLayerPtr spriteLayerP,
	SpritePtr startSpriteP,
	Point testPoint)
{
	SpritePtr curSpriteP;

	curSpriteP = (startSpriteP == NULL) ? spriteLayerP->tailSpriteP : startSpriteP;

		// note that we traverse the sprites in reverse order
		// since this will make the most sense if we are
		// looking for a sprite that was clicked
	while ((curSpriteP != NULL) && !SWIsPointInSprite(curSpriteP, testPoint))
	{
		curSpriteP = curSpriteP->prevSpriteP;
	}

	return curSpriteP;
}
